@@ -11,7 +11,7 @@
  *   for the specific language governing rights and limitations under the
  *   License.
  *
- *   Copyright 2009-2011, Red Hat Inc.
+ *   Copyright 2009-2013, Red Hat Inc.
  *   Based on mozilla.org's scriptable plugin example
  *
  *   The Original Code is mozilla.org code.
@@ -25,6 +25,7 @@
  *   Uri Lublin
  *   Martin Stransky
  *   Peter Hatina
+ *   Christophe Fergeau
  *
  *   Alternatively, the contents of this file may be used under the terms of
  *   either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -44,47 +45,33 @@
 #define _GNU_SOURCE
 #endif
 
-#include <sys/types.h>
-#include <sys/wait.h>
-#include <sys/stat.h>
+#include "config.h"
 
 #include <stdlib.h>
-#include <dlfcn.h>
 #include <errno.h>
 #include <unistd.h>
 #include <string>
 #include <sstream>
 #include <signal.h>
 #include <glib.h>
+#include <glib/gstdio.h>
+#include <gio/gio.h>
 
 extern "C" {
 #include <pthread.h>
 #include <signal.h>
 }
 
-#include "nsCOMPtr.h"
-
-// for plugins
-#ifndef XP_UNIX
-#include <nsString.h>
-#include <nsIDOMNavigator.h>
-#include <nsIDOMPluginArray.h>
-#include <nsIDOMPlugin.h>
-#include <nsIDOMMimeType.h>
-#include <nsIObserverService.h>
-#include <plugin/nsIPluginHost.h>
-
-static NS_DEFINE_CID(kPluginManagerCID, NS_PLUGINMANAGER_CID);
-#endif
-
-#include <nsIServiceManager.h>
-#include <nsISupportsUtils.h> // some usefule macros are defined here
-
+#include <cstring>
 #include <fstream>
 #include <set>
 
-#include "config.h"
-#include "controller.h"
+#if defined(XP_UNIX)
+#include "controller-unix.h"
+#endif
+#if defined(XP_WIN)
+#include "controller-win.h"
+#endif
 #include "plugin.h"
 #include "nsScriptablePeer.h"
 
@@ -120,17 +107,10 @@ namespace {
     }
 }
 
-#ifdef NPAPI_USE_CONSTCHARS
 const char *NPP_GetMIMEDescription(void)
 {
     return const_cast<char *>(MIME_TYPES_DESCRIPTION.c_str());
 }
-#else
-char *NPP_GetMIMEDescription(void)
-{
-    return strdup(MIME_TYPES_DESCRIPTION.c_str());
-}
-#endif
 
 //////////////////////////////////////
 //
@@ -193,25 +173,78 @@ void NS_DestroyPluginInstance(nsPluginInstanceBase *aPlugin)
 //
 // nsPluginInstance class implementation
 //
+static void glib_log_to_file(const gchar *log_domain,
+                             GLogLevelFlags log_level,
+                             const gchar *message,
+                             gpointer user_data)
+{
+    FILE *log_file;
+
+    if ((log_level & G_LOG_LEVEL_MASK) > G_LOG_LEVEL_MESSAGE) {
+        return;
+    }
+
+    log_file = (FILE *)user_data;
+
+    if (log_domain != NULL) {
+        fwrite(log_domain, strlen(log_domain), 1, log_file);
+        fwrite(": ", 2, 1, log_file);
+    }
+    if (message != NULL) {
+        fwrite(message, strlen(message), 1, log_file);
+    }
+    fwrite("\r\n", 2, 1, log_file);
+    fflush(log_file);
+}
+
+static void glib_setup_logging(void)
+{
+#if defined(XP_WIN)
+    FILE *log_file;
+    gchar *log_filename;
+
+    if (!g_getenv("SPICE_XPI_LOG_TO_FILE"))
+        return;
+
+    log_filename = g_build_filename(g_get_tmp_dir(), "SPICEXPI.LOG", NULL);
+    log_file = fopen(log_filename, "w+");
+    if (log_file != NULL) {
+        g_log_set_default_handler(glib_log_to_file, log_file);
+    } else {
+        gchar *log_msg;
+        log_msg = g_strdup_printf("failed to open %s", log_filename);
+        g_free(log_msg);
+    }
+    g_free(log_filename);
+#endif
+}
 
 nsPluginInstance::nsPluginInstance(NPP aInstance):
     nsPluginInstanceBase(),
-    m_pid_controller(-1),
     m_connected_status(-2),
     m_instance(aInstance),
-    m_initialized(PR_TRUE),
+    m_initialized(true),
     m_window(NULL),
-    m_fullscreen(PR_FALSE),
-    m_smartcard(PR_FALSE),
-    m_admin_console(PR_FALSE),
-    m_no_taskmgr_execution(PR_FALSE),
-    m_send_ctrlaltdel(PR_TRUE),
-    m_usb_auto_share(PR_TRUE),
+    m_fullscreen(false),
+    m_smartcard(false),
+    m_admin_console(false),
+    m_no_taskmgr_execution(false),
+    m_send_ctrlaltdel(true),
+    m_usb_auto_share(true),
     m_scriptable_peer(NULL)
 {
-    // create temporary directory in /tmp
-    char tmp_dir[] = "/tmp/spicec-XXXXXX";
-    m_tmp_dir = mkdtemp(tmp_dir);
+#if !GLIB_CHECK_VERSION(2, 35, 0)
+    g_type_init();
+#endif
+    glib_setup_logging();
+
+#if defined(XP_WIN)
+    m_external_controller = new SpiceControllerWin(this);
+#elif defined(XP_UNIX)
+    m_external_controller = new SpiceControllerUnix(this);
+#else
+#error "Unknown OS, no controller implementation"
+#endif
 }
 
 nsPluginInstance::~nsPluginInstance()
@@ -222,14 +255,12 @@ nsPluginInstance::~nsPluginInstance()
     // and zero its m_plugin member
     if (m_scriptable_peer)
         NPN_ReleaseObject(m_scriptable_peer);
-
-    // delete the temporary directory used for a client socket
-    rmdir(m_tmp_dir.c_str());
+    delete(m_external_controller);
 }
 
 NPBool nsPluginInstance::init(NPWindow *aWindow)
 {
-    m_initialized = PR_TRUE;
+    m_initialized = true;
 
     m_host_ip.clear();
     m_port.clear();
@@ -249,12 +280,14 @@ NPBool nsPluginInstance::init(NPWindow *aWindow)
     m_trust_store_file.clear();
     m_color_depth.clear();
     m_disable_effects.clear();
+    m_proxy.clear();
+    m_external_controller->SetProxy(std::string());
 
-    m_fullscreen = PR_FALSE;
-    m_smartcard = PR_FALSE;
-    m_admin_console = PR_FALSE;
-    m_no_taskmgr_execution = PR_FALSE;
-    m_send_ctrlaltdel = PR_TRUE;
+    m_fullscreen = false;
+    m_smartcard = false;
+    m_admin_console = false;
+    m_no_taskmgr_execution = false;
+    m_send_ctrlaltdel = true;
 
     return m_initialized;
 }
@@ -268,7 +301,7 @@ NPError nsPluginInstance::SetWindow(NPWindow *aWindow)
 
 void nsPluginInstance::shut()
 {
-    m_initialized = PR_FALSE;
+    m_initialized = false;
 }
 
 NPBool nsPluginInstance::isInitialized()
@@ -386,23 +419,23 @@ void nsPluginInstance::SetHostSubject(const char *aHostSubject)
 }
 
 /* attribute boolean fullScreen; */
-PRBool nsPluginInstance::GetFullScreen() const
+bool nsPluginInstance::GetFullScreen() const
 {
     return m_fullscreen;
 }
 
-void nsPluginInstance::SetFullScreen(PRBool aFullScreen)
+void nsPluginInstance::SetFullScreen(bool aFullScreen)
 {
     m_fullscreen = aFullScreen;
 }
 
 /* attribute boolean Smartcard; */
-PRBool nsPluginInstance::GetSmartcard() const
+bool nsPluginInstance::GetSmartcard() const
 {
     return m_smartcard;
 }
 
-void nsPluginInstance::SetSmartcard(PRBool aSmartcard)
+void nsPluginInstance::SetSmartcard(bool aSmartcard)
 {
     m_smartcard = aSmartcard;
 }
@@ -441,12 +474,12 @@ void nsPluginInstance::SetNumberOfMonitors(const char *aNumberOfMonitors)
 }
 
 /* attribute boolean AdminConsole; */
-PRBool nsPluginInstance::GetAdminConsole() const
+bool nsPluginInstance::GetAdminConsole() const
 {
     return m_admin_console;
 }
 
-void nsPluginInstance::SetAdminConsole(PRBool aAdminConsole)
+void nsPluginInstance::SetAdminConsole(bool aAdminConsole)
 {
     m_admin_console = aAdminConsole;
 }
@@ -474,23 +507,23 @@ void nsPluginInstance::SetHotKeys(const char *aHotKeys)
 }
 
 /* attribute boolean NoTaskMgrExecution; */
-PRBool nsPluginInstance::GetNoTaskMgrExecution() const
+bool nsPluginInstance::GetNoTaskMgrExecution() const
 {
     return m_no_taskmgr_execution;
 }
 
-void nsPluginInstance::SetNoTaskMgrExecution(PRBool aNoTaskMgrExecution)
+void nsPluginInstance::SetNoTaskMgrExecution(bool aNoTaskMgrExecution)
 {
     m_no_taskmgr_execution = aNoTaskMgrExecution;
 }
 
 /* attribute boolean SendCtrlAltDelete; */
-PRBool nsPluginInstance::GetSendCtrlAltDelete() const
+bool nsPluginInstance::GetSendCtrlAltDelete() const
 {
     return m_send_ctrlaltdel;
 }
 
-void nsPluginInstance::SetSendCtrlAltDelete(PRBool aSendCtrlAltDelete)
+void nsPluginInstance::SetSendCtrlAltDelete(bool aSendCtrlAltDelete)
 {
     m_send_ctrlaltdel = aSendCtrlAltDelete;
 }
@@ -512,12 +545,12 @@ void nsPluginInstance::SetUsbListenPort(unsigned short aUsbPort)
 }
 
 /* attribute boolean UsbAutoShare; */
-PRBool nsPluginInstance::GetUsbAutoShare() const
+bool nsPluginInstance::GetUsbAutoShare() const
 {
     return m_usb_auto_share;
 }
 
-void nsPluginInstance::SetUsbAutoShare(PRBool aUsbAutoShare)
+void nsPluginInstance::SetUsbAutoShare(bool aUsbAutoShare)
 {
     m_usb_auto_share = aUsbAutoShare;
 }
@@ -544,9 +577,21 @@ void nsPluginInstance::SetDisableEffects(const char *aDisableEffects)
     m_disable_effects = aDisableEffects;
 }
 
+/* attribute string Proxy; */
+char *nsPluginInstance::GetProxy() const
+{
+    return stringCopy(m_proxy);
+}
+
+void nsPluginInstance::SetProxy(const char *aProxy)
+{
+    m_proxy = aProxy;
+    m_external_controller->SetProxy(m_proxy);
+}
+
 void nsPluginInstance::WriteToPipe(const void *data, uint32_t size)
 {
-    m_external_controller.Write(data, size);
+    m_external_controller->Write(data, size);
 }
 
 void nsPluginInstance::SendInit()
@@ -591,6 +636,43 @@ void nsPluginInstance::SendStr(uint32_t id, std::string str)
     free(msg);
 }
 
+bool nsPluginInstance::CreateTrustStoreFile(const std::string &trust_store)
+{
+    GFile *tmp_file;
+    GFileIOStream *iostream;
+    GOutputStream *stream;
+
+    tmp_file = g_file_new_tmp("trustore.pem-XXXXXX", &iostream, NULL);
+    if (tmp_file == NULL) {
+        g_message("Couldn't create truststore");
+        return false;
+    }
+
+    stream = g_io_stream_get_output_stream(G_IO_STREAM(iostream));
+    if (!g_output_stream_write_all(stream,
+                                   trust_store.c_str(),
+                                   trust_store.length(),
+                                   NULL, NULL, NULL)) {
+        g_critical("Couldn't write truststore");
+        return false;
+    }
+    m_trust_store_file = g_file_get_path(tmp_file);
+    g_object_unref(tmp_file);
+    g_object_unref(iostream);
+
+    return true;
+}
+
+bool nsPluginInstance::RemoveTrustStoreFile()
+{
+    if (g_unlink(m_trust_store_file.c_str()) != 0)
+        return false;;
+
+    m_trust_store_file.clear();
+
+    return true;
+}
+
 void nsPluginInstance::Connect()
 {
     const int port = portToInt(m_port);
@@ -606,129 +688,49 @@ void nsPluginInstance::Connect()
         return;
     }
 
-    std::string socket_file(m_tmp_dir);
-    socket_file += "/spice-xpi";
-    if (setenv("SPICE_XPI_SOCKET", socket_file.c_str(), 1))
-    {
-        g_critical("could not set SPICE_XPI_SOCKET env variable");
+    if (!m_external_controller->StartClient()) {
+        g_critical("failed to start SPICE client");
         return;
     }
 
-    /* use a pipe for the children to wait until it gets tracked */
-    int pipe_fds[2] = { -1, -1 };
-    if (pipe(pipe_fds) < 0) {
-        perror("spice-xpi system error");
+    if (m_external_controller->Connect(10) != 0)
+    {
+        g_critical("could not connect to spice client controller");
         return;
     }
 
-    m_pid_controller = fork();
-    if (m_pid_controller == 0)
-    {
-        setpgrp();
-
-        close(pipe_fds[1]);
-        pipe_fds[1] = -1;
-
-        char c;
-        if (read(pipe_fds[0], &c, 1) != 0)
-            g_critical("Error while reading on pipe: %s", g_strerror(errno));
-
-        close(pipe_fds[0]);
-        pipe_fds[0] = -1;
-
-        execl("/usr/libexec/spice-xpi-client", "/usr/libexec/spice-xpi-client", NULL);
-        g_message("failed to run spice-xpi-client, running spicec instead");
-
-        // TODO: temporary fallback for backward compatibility
-        execl("/usr/bin/spicec", "/usr/bin/spicec", "--controller", NULL);
-        g_critical("ERROR failed to run spicec fallback");
-
-        exit(EXIT_FAILURE);
+    if (!this->CreateTrustStoreFile(m_trust_store)) {
+        g_critical("failed to create trust store");
+        return;
     }
-    else
-    {
-        g_debug("child pid: %"G_GUINT64_FORMAT, (guint64)m_pid_controller);
 
-        close(pipe_fds[0]);
-        pipe_fds[0] = -1;
+    SendInit();
+    SendStr(CONTROLLER_HOST, m_host_ip);
+    if (port > 0)
+        SendValue(CONTROLLER_PORT, port);
+    if (sport > 0)
+        SendValue(CONTROLLER_SPORT, sport);
+    SendValue(CONTROLLER_FULL_SCREEN,
+            (m_fullscreen == true ? CONTROLLER_SET_FULL_SCREEN : 0) |
+            (m_admin_console == false ? CONTROLLER_AUTO_DISPLAY_RES : 0));
+    SendBool(CONTROLLER_ENABLE_SMARTCARD, m_smartcard);
+    SendStr(CONTROLLER_PASSWORD, m_password);
+    SendStr(CONTROLLER_TLS_CIPHERS, m_cipher_suite);
+    SendStr(CONTROLLER_SET_TITLE, m_title);
+    SendBool(CONTROLLER_SEND_CAD, m_send_ctrlaltdel);
+    SendBool(CONTROLLER_ENABLE_USB_AUTOSHARE, m_usb_auto_share);
+    SendStr(CONTROLLER_USB_FILTER, m_usb_filter);
+    SendStr(CONTROLLER_SECURE_CHANNELS, m_ssl_channels);
+    SendStr(CONTROLLER_CA_FILE, m_trust_store_file);
+    SendStr(CONTROLLER_HOST_SUBJECT, m_host_subject);
+    SendStr(CONTROLLER_HOTKEYS, m_hot_keys);
+    SendValue(CONTROLLER_COLOR_DEPTH, atoi(m_color_depth.c_str()));
+    SendStr(CONTROLLER_DISABLE_EFFECTS, m_disable_effects);
+    SendMsg(CONTROLLER_CONNECT);
+    SendMsg(CONTROLLER_SHOW);
 
-        pthread_t controller_thread_id;
-        pthread_create(&controller_thread_id, NULL, ControllerWaitHelper,
-            reinterpret_cast<void*>(this));
-
-        close(pipe_fds[1]);
-        pipe_fds[1] = -1;
-
-        m_external_controller.SetFilename(socket_file);
-
-        if (m_external_controller.Connect(10) != 0)
-        {
-            g_critical("could not connect to spice client controller");
-            return;
-        }
-
-        // create trust store filename
-        FILE *fp;
-        int fd = -1;
-        char trust_store_template[] = "/tmp/truststore.pem-XXXXXX";
-        mode_t prev_umask = umask(0177);
-        fd = mkstemp(trust_store_template);
-        umask(prev_umask);
-        m_trust_store_file = trust_store_template;
-
-        if (fd != -1)
-        {
-            fp = fdopen(fd,"w+");
-            if (fp != NULL)
-            {
-                fputs(m_trust_store.c_str(), fp);
-                fflush(fp);
-                fsync(fd);
-                fclose(fp);
-            }
-            else
-            {
-                g_critical("could not open truststore temp file");
-                close(fd);
-                unlink(m_trust_store_file.c_str());
-                m_trust_store_file.clear();
-                return;
-            }
-        }
-        else
-        {
-            g_critical("could not create truststore temp file: %s", g_strerror(errno));
-            return;
-        }
-
-        SendInit();
-        SendStr(CONTROLLER_HOST, m_host_ip);
-        if (port > 0)
-            SendValue(CONTROLLER_PORT, port);
-        if (sport > 0)
-            SendValue(CONTROLLER_SPORT, sport);
-        SendValue(CONTROLLER_FULL_SCREEN,
-                   (m_fullscreen == PR_TRUE ? CONTROLLER_SET_FULL_SCREEN : 0) |
-                   (m_admin_console == PR_FALSE ? CONTROLLER_AUTO_DISPLAY_RES : 0));
-        SendBool(CONTROLLER_ENABLE_SMARTCARD, m_smartcard);
-        SendStr(CONTROLLER_PASSWORD, m_password);
-        SendStr(CONTROLLER_TLS_CIPHERS, m_cipher_suite);
-        SendStr(CONTROLLER_SET_TITLE, m_title);
-        SendBool(CONTROLLER_SEND_CAD, m_send_ctrlaltdel);
-        SendBool(CONTROLLER_ENABLE_USB_AUTOSHARE, m_usb_auto_share);
-        SendStr(CONTROLLER_USB_FILTER, m_usb_filter);
-        SendStr(CONTROLLER_SECURE_CHANNELS, m_ssl_channels);
-        SendStr(CONTROLLER_CA_FILE, m_trust_store_file);
-        SendStr(CONTROLLER_HOST_SUBJECT, m_host_subject);
-        SendStr(CONTROLLER_HOTKEYS, m_hot_keys);
-        SendValue(CONTROLLER_COLOR_DEPTH, atoi(m_color_depth.c_str()));
-        SendStr(CONTROLLER_DISABLE_EFFECTS, m_disable_effects);
-        SendMsg(CONTROLLER_CONNECT);
-        SendMsg(CONTROLLER_SHOW);
-
-        // set connected status
-        m_connected_status = -1;
-    }
+    // set connected status
+    m_connected_status = -1;
 }
 
 void nsPluginInstance::Show()
@@ -739,11 +741,10 @@ void nsPluginInstance::Show()
 
 void nsPluginInstance::Disconnect()
 {
-    if (m_pid_controller > 0)
-        kill(-m_pid_controller, SIGTERM);
+    m_external_controller->StopClient();
 }
 
-void nsPluginInstance::ConnectedStatus(PRInt32 *retval)
+void nsPluginInstance::ConnectedStatus(int32_t *retval)
 {
     *retval = m_connected_status;
 }
@@ -812,27 +813,16 @@ void nsPluginInstance::CallOnDisconnected(int code)
     NPN_ReleaseVariantValue(&var_on_disconnected);
 }
 
-void *nsPluginInstance::ControllerWaitHelper(void *opaque)
+void nsPluginInstance::OnSpiceClientExit(int exit_code)
 {
-    nsPluginInstance *fake_this = reinterpret_cast<nsPluginInstance *>(opaque);
-    if (!fake_this)
-        return NULL;
-
-    int exit_code;
-    waitpid(fake_this->m_pid_controller, &exit_code, 0);
-    g_debug("child finished, pid: %"G_GUINT64_FORMAT, (guint64)exit_code);
-
-    fake_this->m_connected_status = fake_this->m_external_controller.TranslateRC(exit_code);
+    m_connected_status = m_external_controller->TranslateRC(exit_code);
     if (!getenv("SPICE_XPI_DEBUG"))
     {
-        fake_this->CallOnDisconnected(exit_code);
-        fake_this->m_external_controller.Disconnect();
+        CallOnDisconnected(exit_code);
+        m_external_controller->Disconnect();
     }
-    
-    unlink(fake_this->m_trust_store_file.c_str());
-    fake_this->m_trust_store_file.clear();
-    fake_this->m_pid_controller = -1;
-    return NULL;
+
+    RemoveTrustStoreFile();
 }
 
 // ==============================
